#!/usr/bin/perl -w 

use strict;
use URI::Escape;
use HTML::Entities;

######################################################################

sub usage {
    print STDERR "Usage:\n  $0 [ -expecterror errcode url ] [ -expectredirect fromurl tourl ] [ -multi mainemail\@host [ -subtree url email\@host ]* ] references statuses\n";
}

######################################################################

sub uniq {
    my (@rval, %uniq);
    local ($_);
    foreach $_ (@_) {
	next if $uniq{$_};
	push(@rval, $_);
	$uniq{$_} = 1;
    }
    return @rval;
}


######################################################################

package badURLReport;

sub new {
    shift;

    my ($url) = @_;

    unless (defined $::referees{$url}) {
	print STDERR "no referee $url\n";
    }
    my (@bad_referers) = ::uniq @{$::referees{$url}};

    my ($self) = { "url" => $url ,
		   "referers" => \@bad_referers };
    
    bless $self;

    return $self;
}

sub classifierFor {
    my ($self) = shift;
    my ($email) = @_;

    my ($subroutine);
    if (defined $email && defined $::mainemail) {
	my ($pattern);
	if ($email eq "_") {
	    my ($p2, @urls);
	    @urls = map { ($_ eq "")?():@{$::reportto{$_}} } (keys %::reportto);
	    foreach $_ (@urls) { s/(\W)/\\$1/g; };
	    $pattern = "^(" . join("|", @urls) . ")";

	    if (defined $::reportto{$::mainemail}) {
		@urls = @{ $::reportto{$::mainemail} };
		foreach $_ (@urls) { s/(\W)/\\$1/g; };
		$p2 = "^(" . join("|", @urls) . ")";
		return eval "sub { local \$_ = shift; return /$p2/o || ! /$pattern/o; }";
	    } else {
		return eval "sub { local \$_ = shift; return !/$pattern/o; }";
	    }
	} else {
	    my (@urls) = @{ $::reportto{$email} };

	    foreach $_ (@urls) { s/(\W)/\\$1/g; }
	    $pattern = "^(" . join("|", @urls) . ")";
	    return eval "sub { local \$_ = shift; return /$pattern/o; }";
	}
    } else {
	return eval "sub { return 1 }";
    }

    return undef;
}

sub reportFor {
    my ($self) = shift;
    my ($email) = @_;

    my ($classifier) = $self->classifierFor($email);

    my ($url, $all_bad_referers) = ($self->{"url"}, $self->{"referers"});
    my (%bad_referers, $rval, $anyFound, $stat);

    $stat = $::statuses{$url};
    my ($urlD, $urlE, $urlH) = ::assorted($url);

    $rval =  "<p>$stat <a href=\"$urlE\">$urlH</a>";
    if (defined $::redirections{$url}) {
	my $redir = $::redirections{$url};
	my ($rD, $rE, $rH) = ::assorted($redir);
	$rval .= "\n<br>redirected to <a href=\"$rE\">$rH</a><br>\n"
	}
    $rval .= " referred to by\n<ul>\n";

    my ($refD, $refE, $refH);
    foreach (@{$all_bad_referers}) {
	next unless &$classifier($_);

	$anyFound = 1;
	$bad_referers{$_} = 1;
	($refD, $refE, $refH) = ::assorted($_);
	$rval .= "  <li> <a href=\"$refE\">$refH</a> (<a href=\"#doc_$refE\">report</a>)\n";
    }

    return () unless $anyFound;

    $rval .= "</ul>\n\n";

    return ($rval, keys %bad_referers);
}

package main;

######################################################################

# load a file full of "referrer referee" lines which tell you which
# documents referenced which URLs.  Both URLs are %-encoded, probably
# too aggressively.  We'll decode them now to save confusion later.
#
#  The data is placed into @::reflist, %::referenced, and %::referees.
#
sub load_references {
    my ($fname) = @_;
    open(FILE, "<$fname") || die "Unable to open $fname for read";
    my ($doc, $ref);
    local ($_);

    @::reflist = ();
    %::referenced = ();
    %::referees = ();

    while (<FILE>) {
	chomp;
	($doc, $ref) = split(/\s+/, $_, 2);

	$doc = uri_unescape($doc);
	$ref = uri_unescape($ref);

	# code for this oddity later:
	$ref =~ s/#.*//;

	push(@::reflist, [$doc, $ref] );

	$::referenced{$doc} = [] unless defined $::referenced{$doc};
	$::referenced{$doc} = [ @{$::referenced{$doc}}, $ref ];

	($::referees{$ref} = []) unless (defined $::referees{$ref});
	$::referees{$ref} = [ @{$::referees{$ref}}, $doc ];

    }
    close(FILE);
}


# load a file full of "status url"
#
#  The data is placed in %::statuses and %::redirections
#
sub load_statuses {
    my ($fname) = @_;
    open(FILE, "<$fname") || die "Unable to open $fname for read";
    my ($stat, $doc, $redir);
    local ($_);

    %::statuses = ();
    %::redirections = ();

    while (<FILE>) {
	chomp;
	($stat, $doc) = split(/\s+/, $_, 2);

	$doc = uri_unescape($doc);

	if ($stat =~ /^30[127]$/) {
	    ($doc, $redir) = split(/\s+/, $doc, 2);
	    $::redirections{$doc} = $redir;
	}
	$::statuses{$doc} = $stat;
    }
}

sub bad_stat
{
    my ($urlD, $stat) = @_;

    return 0 if ($stat =~ /^2\d\d$/);

    return 0 if ($stat eq "badProtocol" &&
		 $urlD =~ /^(ftp|mailto):/);

    return 1;
}

sub pick_bad_refs
{
    my ($XPCTerr, $XPCTredir) = @_;

    my ($url, $urlD, $stat, @rval);

    while ( ($url, $stat) = each %::statuses ) {
	next unless bad_stat($url, $stat);
	next if (defined($XPCTerr->{$url}) &&
		 $XPCTerr->{$url} eq $stat);
	next if (defined($::redirections{$url}) &&
		 defined($XPCTredir->{$url}) &&
		 $::redirections{$url} eq $XPCTredir->{$url});

	push(@rval, $url);
    }

    return @rval;
}

sub cmp_status
{
    my $x;
    $x = $::statuses{$a} cmp $::statuses{$b};
    return $x if $x;
    return $a cmp $b;
}

sub assorted
{
    my ($d) = shift;
    return ($d, uri_escape($d), encode_entities($d));
}

# use the "select FILEHANDLE" operator to control where this is output
sub begin_report {
    print <<EOF
<html>
<title>link checker report</title>
</html>
<body>
<h1>link checker report</h1>

<h2> bad referenced URLs</h2>
EOF
    ;
}

sub report_failed_references {
    my ($fhr) = shift;

    my (@bad) = sort cmp_status @_;

    my (%referer_bad, $url, @emails, $email, $fh);

    local $_;

    if (defined $fhr) {
	@emails = keys %{$fhr};
    } else {
	@emails = "_";
    }

    foreach $email (@emails) { $referer_bad{$email}  = []; }

    foreach $url (@bad) {
	$_ = new badURLReport($url);
	foreach $email (@emails) {
	    my ($txt, @bad) = $_->reportFor($email);
	    next unless defined $txt;
	    $fh = $fhr->{$email};
	    { no strict "refs";
	      print $fh $txt; }
	    push @{ $referer_bad{$email} }, @bad;
	}
    }

    return %referer_bad;
}

sub per_referer_report {
    my ($url, $XPCTerr, $XPCTredir) = @_;
    my ($rval,
	$urlD, $urlE, $urlH,
	$stat,
	$ref, $refD, $refE, $refH);

#    print STDERR join(":", keys %{$XPCTredir}), "\n";

    $stat = $::statuses{$url};
    ($urlD, $urlE, $urlH) = assorted($url);
    $rval = "<p> <a name=\"doc_$urlE\">$stat</a> <a href=\"$urlE\">$urlH</a>";

    if (defined $::redirections{$url}) {
	my $redir = $::redirections{$url};
	my ($rD, $rE, $rH) = assorted($redir);
	$rval .= "\n<br>redirected to <a href=\"$rE\">$rH</a><br>\n";
    }

    $rval .= " references:\n<ul>\n";
    foreach $ref (@{$::referenced{$url}}) {
	($refD, $refE, $refH) = assorted($ref);
	$stat = $::statuses{$ref};
	next unless defined($stat) && bad_stat($refD, $stat);
	next if (defined($XPCTerr->{$refD}) &&
		 $XPCTerr->{$refD} eq $stat);
	next if (defined($::redirections{$refD}) &&
		 defined($XPCTredir->{$refD}) &&
		 $::redirections{$refD} eq $XPCTredir->{$refD});
	if (defined($XPCTredir->{$refD})) {
	    print STDERR $refD, " -> ", $XPCTredir->{$refD}, "\n";
	}
	$rval .= "  <li> $stat <a href=\"$refE\">$refH</a>";
#	    print " (<a href=\"#$refE\">report</a>)";
	$rval .= "\n";
    }
    $rval .= "</ul>\n\n";

    return $rval;
}

sub report_documents
{
    my ($fhr, $referer_bad, $XPCTerr, $XPCTredir) = (@_);
    my (@docs, %cache, $fh, $url, $email);
    local $_;


    my ($oldfh) = select;

    foreach $email (keys %{$fhr}) {
	$fh = $fhr->{$email};
	select $fh;

	print "<h2> documents that reference bad links</h2>\n\n";

	@docs = sort { $a cmp $b } uniq @{$referer_bad->{$email}};
	foreach $url (@docs) {
	    unless (defined $cache{$url})  {
		$cache{$url} = per_referer_report($url, $XPCTerr, $XPCTredir);
	    }

	    no strict "refs";
	    print $cache{$url};
	}
    }

    select $oldfh;
}

######################################################################
#
######################################################################

my @include_exclude = ();

while (@ARGV) {
    local $_ = shift;
    if ( $_ eq "-excludetree") {
	my $x = shift;
	push (@include_exclude, [ $x, 0, 1 ]);
    } else {
	unshift(@ARGV, $_);
	last;
    }
}

my (%expecterror, %expectredirect);

while (@ARGV) {
    local ($_) = shift;

    if ($_ eq "-expecterror") {
	if (@ARGV<2) { usage(); exit 1; }
	my ($errcode)  = shift;
	my ($url) = shift;
	$url = uri_unescape($url);
	$expecterror{$url} = $errcode;
    } elsif ($_ eq "-expectredirect") {
	if (@ARGV<2) { usage(); exit 1; }
	my ($fromurl)  = shift;
	$fromurl = uri_unescape($fromurl);
	my ($tourl) = shift;
	$tourl = uri_unescape($tourl);
	$expectredirect{$fromurl} = $tourl;
    } elsif ($_ eq "-multi") {
	if (@ARGV<1) { usage(); exit 1; }
	$::mainemail = shift;
    } elsif ($_ eq "-subtree") {
	if (!defined $::mainemail) {
	    warn "-subtree before -multi not permitted";
	    usage();
	    exit 1;
	}
	if (@ARGV<2) { usage(); exit 1; }
	my ($url ) = shift;
	$url = uri_unescape($url);
	my ($email ) = shift;
	if (defined $::reportto{$email}) {
	    push (@{$::reportto{$email}}, $url);
	} else {
	    $::reportto{$email} = [ $url ];
	}
    } else {
	unshift(@ARGV, $_);
	last;
    }
}

if (@ARGV!=2) {
    usage();
    exit 1;
}

my ($ref_fname, $stat_fname) = @ARGV;

&load_references($ref_fname);
&load_statuses($stat_fname);
# The following objects are now populated:
# all URLs are %-escaped
# @::reflist	(array of "referer referee" []s)
# %::referenced	(mapping from document to [] of referenced URLs )
# %::referees	(mapping from URLs to [] of documents that reference them)
# %::statuses	(mapping from URLs to result of status probe)
# %::redirections

#print STDERR join(",", @{$::referees{"http%3A%2F%2Fwww.clas.ufl.edu%2Fgermslav%2F"}}), "\n";

my (@failed_references) = &pick_bad_refs(\%expecterror, \%expectredirect);
# all URLs are %-escaped

my ($fhr);

if (defined $::mainemail) {
    my ($fh) = "FILEAAA0";
    my ($fname, $email);

    $fname = "report-$::mainemail.html";
    {
	no strict "refs";
	open($fh, ">$fname") || die "Unable to open $fname for write";
    }
    $fhr->{"_"} = $fh;
    select $fh;
    begin_report();

    foreach $email (keys %::reportto) {
	next if $email eq $::mainemail;
	$fh++;
	$fname = "report-$email.html";
	{
	    no strict "refs";
	    open($fh, ">$fname") || die "Unable to open $fname for write";
	}
	$fhr->{$email} = $fh;
	select $fh;
	begin_report();
    }
    select STDOUT;
} else {
    begin_report();
    $fhr->{"_"} = "STDOUT";
}

my (%flawed_documents) = &report_failed_references( $fhr,
						   @failed_references);

&report_documents($fhr, \%flawed_documents, \%expecterror, \%expectredirect);

if (defined $::mainemail) {
    my ($email);
    foreach $email (keys %{$fhr}) {
	no strict "refs";
	close($fhr->{$email});
    }
}

exit 0;
